
/* Copyright (c) 2007, Matt Parker
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the <organization> nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY Matt Parker ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Matt Parker BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "client.h"
#include "quake2ai.h"

trace_t CL_Trace (vec3_t start, vec3_t end, float size, int contentmask);
trace_t CL_Trace (vec3_t start, vec3_t end, float size,  int contentmask)
{
	vec3_t maxs, mins;

	VectorSet(maxs, size, size, size);
	VectorSet(mins, -size, -size, -size);

	return CM_BoxTrace (start, end, mins, maxs, 0, contentmask);
}

void AI_DoMovements (usercmd_t *cmd);

extern void IN_AttackDown(void);
extern void IN_AttackUp(void);

entity_t *AI_entities;

/* redefine AImain to do ai, and use callback for threads*/
extern void AImain (void);
extern void callback (void);
void (*AImain_callback)(void) = NULL;
void (*AImain_callback2)(void) = NULL;


struct AIself_t {
	float forwardspeed;
	float sidespeed;
	float upspeed;
	double pitchspeed;
	double yawspeed;
	double pitch;
	double yaw;
	int setpitch;
	int setyaw;
	int attack;
	int attackchange;
} AIself;

#define AI_MAXMSG	128

struct AImsg_t {
	char *msg;
	int type;
} AImsg[AI_MAXMSG];

void AI_setcallback(void (*func)()) {
	AImain_callback = func;
	return;
}

void AI_setcallback2(void (*func)()) {
	AImain_callback2 = func;
	return;
}

void AI_init(void) {
	int i;
	AI_grabmouse = 0;
	AIself.forwardspeed = 0;
	AIself.sidespeed = 0;
	AIself.upspeed = 0;
	AIself.pitchspeed = 0;
	AIself.yawspeed = 0;
	AIself.pitch = 0;
	AIself.setpitch = 0;
	AIself.setyaw = 0;
	AIself.attack = 0;
	AIself.attackchange = 0;
	
	i = 0;
	while (i < AI_MAXMSG) {
		AImsg[i].msg = NULL;
		AImsg[i].type = -1;
		i++;
	}

	i = 0;
	while (i < AI_TALKMAX) {
		AI_talkbank[i] = (char *)malloc(sizeof(char) * AI_MSGLENGTH);
		(AI_talkbank[i])[0] = 0;
		i++;
	}
	AI_talkframes = AI_TALKFRAMES;

	return;
}

int max(int a, int b) {
	if (a > b) return a;
	else return b;
}
int min(int a, int b) {
	if (a < b) return a;
	else return b;
}
double fmax(double a, double b) {
	if (a > b) return a;
	else return b;
}
double fmin(double a, double b) {
	if (a < b) return a;
	else return b;
}
void AI_main(void) {
	/*refresh variables*/
	AIself.pitch = cl.predicted_angles[PITCH];
	AIself.yaw = cl.predicted_angles[YAW];
	/*end update*/
	
	AImain();
	if (AImain_callback != NULL)
		AImain_callback();

	AI_talkframes = max(AI_talkframes - 1, 0);
	AI_pumptalk();

	if (AImain_callback2 != NULL)
		AImain_callback2();
	
	callback();
	AImsg_clear();

	return;
}

int AIself_health(void) { return cl.frame.playerstate.stats[STAT_HEALTH]; }
int AIself_armor(void) { return cl.frame.playerstate.stats[STAT_ARMOR]; }
int AIself_ammo(void) { return cl.frame.playerstate.stats[STAT_AMMO]; }
int AIself_frags(void) { return cl.frame.playerstate.stats[STAT_FRAGS]; }

double AIself_x(void) { return cl.frame.playerstate.pmove.origin[0] / 8.0; }
double AIself_y(void) { return cl.frame.playerstate.pmove.origin[1] / 8.0; }
double AIself_z(void) { return cl.frame.playerstate.pmove.origin[2] / 8.0; }

double AIself_xvel(void) { return cl.frame.playerstate.pmove.velocity[0] / 8.0; }
double AIself_yvel(void) { return cl.frame.playerstate.pmove.velocity[1] / 8.0; }
double AIself_zvel(void) { return cl.frame.playerstate.pmove.velocity[2] / 8.0; }

void AIself_forwardmove(float value) {
	AIself.forwardspeed = value;
	return;
}

void AIself_sidemove(float value) {
	AIself.sidespeed = value;
	return;
}

void AIself_upmove(float value) {
	AIself.upspeed = value;
	return;
}

void AIself_attack(int value) {
	if (AIself.attack != value) {
		AIself.attackchange = 1;
		AIself.attack = value;
	}
	return;
}

void AIself_yawmove(double angle) {
	AIself.yawspeed = angle;
	return;
}

void AIself_setyaw(double angle) {
	AIself.yaw = angle;
	AIself.setyaw = 1;
	return;
}

double AIself_yaw(void) {
	return cl.predicted_angles[YAW];
}

void AIself_pitchmove(double angle) {
	AIself.pitchspeed = angle;
	return;
}

void AIself_setpitch(double angle) {
	AIself.pitch = angle;
	AIself.setpitch = 1;
	return;
}

double AIself_pitch(void) {
	return cl.predicted_angles[PITCH];
}

trace_t AItrace_tr;

void AItrace (float startx, float starty, float startz, float endx, float endy, float endz, float size) {
	vec3_t startv, endv;
	
	startv[0] = startx;
	startv[1] = starty;
	startv[2] = startz;
	
	endv[0] = endx;
	endv[1] = endy;
	endv[2] = endz;
	
	AItrace_tr = CL_Trace(startv, endv, size, 1);
		
	return;
}

float AItraced_endpoint_x () { return AItrace_tr.endpos[0]; }
float AItraced_endpoint_y () { return AItrace_tr.endpos[1]; }
float AItraced_endpoint_z () { return AItrace_tr.endpos[2]; }
float AItraced_fraction () { return AItrace_tr.fraction; }



int AIentity_exists (int ent_num) {
	if (ent_num < AI_entities_num)
		return 1;
	else
		return 0;
}

char *AIentity_model(int ent_num) {
	return AI_entities[ent_num].model;
}

int AIentity_frame(int ent_num) {
	return AI_entities[ent_num].frame;
}

double AIentity_x(int ent_num) {
	return AI_entities[ent_num].origin[0];
}

double AIentity_y(int ent_num) {
	return AI_entities[ent_num].origin[1];
}

double AIentity_z(int ent_num) {
	return AI_entities[ent_num].origin[2];
}

double AIentity_pitch(int ent_num) {
	return AI_entities[ent_num].angles[PITCH];
}

double AIentity_yaw(int ent_num) {
	return AI_entities[ent_num].angles[YAW];
}


int AIview_width(void) {
	return re.AIscreen_width();
}

int AIview_height(void) {
	return re.AIscreen_height();
}

int AIview_pixel(int x, int y) {
	return re.AIscreen_pixel(x, y);
}


int AIview_block(int x, int y, int width, int height) {
	return re.AIscreen_block(x, y, width, height);
}

char *AIself_name (void) {
	return (cl.clientinfo[cl.playernum]).name;
}



int blue(int color) {
	return 0x0000FF & color;
}

int green(int color) {
	return (0x00FF00 & color) >> 8;
}

int red(int color) {
	return (0xFF0000 & color) >> 16;
}
	
extern char	chat_buffer[256];
extern int chat_bufferlen;

void reorder_talk(void) {
	int i = 0;
	char *firsttalk = AI_talkbank[0];
	
	while (i < (AI_TALKMAX - 1)) {
		AI_talkbank[i] = AI_talkbank[i+1];
		i++;
	}

	AI_talkbank[AI_TALKMAX - 1] = firsttalk;
	(AI_talkbank[AI_TALKMAX - 1])[0] = 0;
	
	return;
}

void AI_talk(char *msg)
{
	int i = 0;
	while (i < AI_TALKMAX && (AI_talkbank[i])[0] != 0)
		i++;
	
	strcpy(AI_talkbank[i], msg);
	AI_pumptalk();
	return;
}

void AI_teamtalk(char *msg)
{
	int i = 0;
	while (i < AI_TALKMAX && (AI_talkbank[i])[0] != 0)
		i++;
	
	strcpy(AI_talkbank[i], msg);
	AI_pumptalk();
	return;
}

void AI_pumptalk(void)
{
	if (AI_talkframes == 0 && (AI_talkbank[0])[0] != 0) {
		AI_sendtalk(AI_talkbank[0]);
		reorder_talk();
		AI_talkframes = AI_TALKFRAMES;
	}
	
	return;	
}


void AI_sendtalk(char *text) {
	Cbuf_AddText ("say \"");

	strcpy(chat_buffer, text);
	chat_bufferlen = strlen(text);
	
	Cbuf_AddText(chat_buffer);
	Cbuf_AddText("\"\n");

	chat_bufferlen = 0;
	chat_buffer[0] = 0;
	
	return;
}

void AI_sendteamtalk(char *text) {
	Cbuf_AddText ("say_team \"");
	
	strcpy(chat_buffer, text);
	chat_bufferlen = strlen(text);
	
	Cbuf_AddText(chat_buffer);
	Cbuf_AddText("\"\n");

	chat_bufferlen = 0;
	chat_buffer[0] = 0;
	
	return;
}

void AI_command(char *cmd) {
	Cbuf_AddText (cmd);
	return;
}

void AImsg_clear(void) {
	int i = 0;
	while (i < AI_MAXMSG && AImsg[i].msg != NULL) {
		AImsg[0].type = -1;
		free(AImsg[0].msg);
		AImsg[0].msg = NULL;
		i++;
	}
	return;
}

char *AImsg_body (int which) {
	return AImsg[which].msg;
}

int AImsg_type (int which) {
	return AImsg[which].type;
}

char *AImsg_parse(char *msg, int type) {
	int i = 0;
	while (i < AI_MAXMSG && AImsg[i].msg != NULL) i++;
	
	if (i < AI_MAXMSG) {
		AImsg[i].msg = (char*)malloc(sizeof(char) * (1 + strlen(msg)));
		strcpy(AImsg[i].msg, msg);
		AImsg[i].type = type;
		i++;
	}
	return msg;
}


void AI_DoMovements (usercmd_t *cmd) {
	AI_entities_num = 0;

	if (AIself.setyaw) {
		cl.viewangles[YAW] = angleadd(cl.viewangles[YAW], anglediff(cl.predicted_angles[YAW], AIself.yaw));
		AIself.setyaw = 0;
	}
	else {
		cl.viewangles[YAW] = angleadd(cl.viewangles[YAW], AIself.yawspeed);
	}
	if (AIself.setpitch) {
		cl.viewangles[PITCH] = AIself.pitch + anglediff(cl.predicted_angles[PITCH], cl.viewangles[PITCH]);
		AIself.setpitch = 0;
	}
	else {
		cl.viewangles[PITCH] += AIself.pitchspeed;
	}

	cmd->forwardmove += AIself.forwardspeed * 400;
	cmd->sidemove += AIself.sidespeed * 400;
	cmd->upmove += AIself.upspeed * 400;

	if (AIself.attackchange) {
		if (AIself.attack) Cbuf_AddText("+attack\n");
		else Cbuf_AddText("-attack\n");
		AIself.attackchange = 0;
	}
	
	AIself.forwardspeed = 0;
}

double anglediff (double angle1, double angle2)
{
	double selfdir;
	double targetdir;
	targetdir = angle2;
	selfdir = angle1;

	if (targetdir - selfdir > 180) 
		selfdir = selfdir + 360;
	else if (targetdir - selfdir < -180)
		selfdir = selfdir - 360;

	return targetdir - selfdir;
}

double angleadd (double angle1, double angle2)
{
	double finaldir;
	finaldir = angle1 + angle2;
	
	if (finaldir > 360)
		while (finaldir > 360) finaldir -= 360;
	else if (finaldir < 0)
		while (finaldir < 0) finaldir += 360;

	return finaldir;
}

double dist (double x1, double y1, double x2, double y2) {
	return pow(pow(abs(x2 - x1), 2) + pow(abs(y2 - y1), 2), 0.5);	
}

double dist3d (double x1, double y1, double z1, double x2, double y2, double z2) {
	return pow(pow(abs(x2 - x1), 2) + pow(abs(y2 - y1), 2) + pow(abs(z2 - z1), 2), 0.5);
}

double xdir(double x1, double y1, double x2, double y2) {
	double distance = dist(x1, y1, x2, y2);
	if (distance == 0 || (y2 - y1) == 0)
		return 0;
	else if ((y2 - y1) < 0)
		return deg(M_PI * 2 - acos(fmax(-1.0, fmin(1.0, (x2 - x1)/distance))));
	else
		return deg(acos(fmax(-1.0, fmin(1.0, (x2 - x1)/distance))));
}

double rad(double degree)
{
	return 2.0 * M_PI / 360.0 * degree;
}

double deg(double radian)
{
	return (360.0 * radian) / (2.0 * M_PI);
}

void AImain (void) { return; }
void callback (void) { return; }
