/* Copyright (c) 2007, Matt Parker
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the <organization> nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY Matt Parker ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Matt Parker BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* set quake2 command line arguments */
void quake2ai_setargs (char *linecommand);

/* launch the game */
int quake2ai_launch (void);
int quake2ai_launch_threaded (void);

/* launches quake2 with argv as command-line arguments and argc arg count */
int quake2ai_main(int argc, char **argv);


/* redefine this function.  It's called every frame */
void AI_main(void);

void AI_setcallback(void (*func)());
void AI_setcallback2(void (*func)());

char *AIself_name (void);

int AIself_health(void);
int AIself_armor(void);
int AIself_ammo(void);
int AIself_frags(void);

double AIself_x(void);
double AIself_y(void);
double AIself_z(void);

double AIself_xvel(void);
double AIself_yvel(void);
double AIself_zvel(void);

void AIself_forwardmove(float value);
void AIself_sidemove(float value);
void AIself_upmove(float value);
void AIself_attack(int value);

void AIself_yawmove(double angle);
void AIself_setyaw(double angle);
double AIself_yaw(void);
void AIself_pitchmove(double angle);
void AIself_setpitch(double angle);
double AIself_pitch(void);

void AItrace (float startx, float starty, float startz, float endx, float endy, float endz, float size);
float AItraced_endpoint_x (void);
float AItraced_endpoint_y (void);
float AItraced_endpoint_z (void);
float AItraced_fraction (void);

int AIentity_exists (int ent_num);
char *AIentity_model(int ent_num);
double AIentity_x(int ent_num);
double AIentity_y(int ent_num);
double AIentity_z(int ent_num);
double AIentity_pitch(int ent_num);
double AIentity_yaw(int ent_num);
int AIentity_frame(int ent_num);

int AIview_width(void);

int AIview_height(void);

/* returns RGB value of pixel [x, y] */
int AIview_pixel(int x, int y);

/* returns average RGB value of a block of pixels */
int AIview_block(int x, int y, int width, int height);

/* get the R, G, or B from the RGB value */
int red(int color);
int green(int color);
int blue(int color);


/* player chat */
void AI_talk(char *msg);
void AI_teamtalk(char *msg);


/* execute a quake2 console command */
void AI_command(char *cmd);


/* get death messages and player chat */
char *AImsg_body (int which);
int AImsg_type (int which);

/* useful math functions */
double anglediff (double angle1, double angle2);
double angleadd (double angle1, double angle2);
double dist(double x1, double y1, double x2, double y2);
double dist3d (double x1, double y1, double z1, double x2, double y2, double z2);
double xdir(double x1, double y1, double x2, double y2);
double rad(double degree);
double deg(double radian);


/* non-user functions */

int AI_entities_num;

char *AImsg_parse(char *msg, int type);

int AI_grabmouse;
void AI_init(void);

#define AI_TALKMAX	128
#define AI_MSGLENGTH	180
#define AI_TALKFRAMES	4
char *AI_talkbank[AI_TALKMAX];
int AI_talkframes;
void reorder_talk(void);
void AI_pumptalk(void);

void AI_sendtalk(char *text);
void AI_sendteamtalk(char *text);

void AImsg_clear(void);
